﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq.Expressions;
using System.Reflection;

namespace INotifyPropertyChangedTests
{
    public class PropertyNameTestClass : INotifyPropertyChanged
    {
        #region Lambda expression pattern

        public static PropertyChangedEventArgs CreateArgs<T>(
          Expression<Func<T, Object>> propertyExpression)
        {
            return new PropertyChangedEventArgs(
                GetPropertyName<T>(propertyExpression));
        }

        public static string GetPropertyName<T>(
            Expression<Func<T, Object>> propertyExpression)
        {
            var lambda = propertyExpression as LambdaExpression;
            MemberExpression memberExpression;
            if (lambda.Body is UnaryExpression)
            {
                var unaryExpression = lambda.Body as UnaryExpression;
                memberExpression = unaryExpression.Operand as MemberExpression;
            }
            else
            {
                memberExpression = lambda.Body as MemberExpression;
            }

            var propertyInfo = memberExpression.Member as PropertyInfo;

            return propertyInfo.Name;
        }

        #endregion

        #region Properties

        // Variante 1
        public int Prop1 { get; set; }

        // Variante 2
        private int prop2;
        public int Prop2
        {
            get { return prop2; }
            set { prop2 = value; }
        }

        // Variante 3: Emptry string
        private int prop3;
        public int Prop3
        {
            get { return prop3; }
            set
            {
                if (prop3 != value)
                {
                    prop3 = value;
                    OnPropertyChanged(String.Empty);
                }
            }
        }

        // Variante 4: Static string
        private int prop4;
        public int Prop4
        {
            get { return prop4; }
            set
            {
                if (prop4 != value)
                {
                    prop4 = value;
                    OnPropertyChanged("Prop4");
                }
            }
        }

        // Variante 5: StackFrame
        private int prop5;
        public int Prop5
        {
            get { return prop5; }
            set
            {
                if (prop5 != value)
                {
                    prop5 = value;
                    OnPropertyChanged();
                }
            }
        }

        // Variante 6: Lambda expression
        private int prop6;
        public int Prop6
        {
            get { return prop6; }
            set
            {
                if (prop6 != value)
                {
                    prop6 = value;
                    OnPropertyChanged(GetPropertyName(x => x.Prop6));
                }
            }
        }

        // Variante 7: GetCurrentMethod()
        private int prop7;
        public int Prop7
        {
            get { return prop7; }
            set
            {
                if (prop7 != value)
                {
                    prop7 = value;
                    OnPropertyChanged(MethodInfo.GetCurrentMethod());
                }
            }
        }

        // Variante 8: Static EventArgs with Lambda expressions
        private static PropertyChangedEventArgs prop8ChangeArgs =
            CreateArgs<PropertyNameTestClass>(x => x.Prop8);

        private int prop8;
        public int Prop8
        {
            get { return prop8; }
            set
            {
                if (prop8 != value)
                {
                    prop8 = value;
                    OnPropertyChanged(prop8ChangeArgs);
                }
            }
        }

        #endregion

        #region Property name helper

        private void OnPropertyChanged(PropertyChangedEventArgs e)
        {
            PropertyChangedEventHandler tempHandler = PropertyChanged;
            if (tempHandler != null)
                tempHandler(this, e);
        }

        private void OnPropertyChanged(String propertyName)
        {
            PropertyChangedEventHandler tempHandler = PropertyChanged;
            if (tempHandler != null)
                tempHandler(this, new PropertyChangedEventArgs(propertyName));
        }

        private void OnPropertyChanged(MethodBase methodBase)
        {
            OnPropertyChanged(methodBase.Name.Substring(4));
        }

        protected String GetPropertyName<T>(Expression<Func<PropertyNameTestClass, T>> projection)
        {
            return PropertyName(projection);
        }

        private static string PropertyName<T>(Expression<Func<PropertyNameTestClass, T>> projection)
        {
            MemberExpression memberExpression = (MemberExpression)projection.Body;
            return memberExpression.Member.Name;
        }

        protected String GetPropertyName()
        {
            return new StackFrame(1).GetMethod().Name.Substring(4);
        }

        #endregion

        #region INotifyPropertyChanged Member

        /// <summary>
        /// Occurr if a property is changed.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Called when [property changed].
        /// </summary>
        protected void OnPropertyChanged()
        {
            MethodBase mb = new StackFrame(1).GetMethod();
            string propertyName = mb.Name.Substring(4);
            OnPropertyChanged(propertyName);
        }

        #endregion
    }
}